/*---------------------------------------------------------------------------

   spamplot.c                               Copyright 2008-2009 Greg Roelofs

     Plot "pixel-perfect" graph of spams-per-day histogram data, i.e., one
     pixel per spam e-mail (vertically) and one pixel per day (horizontally).

                 http://gregroelofs.com/greg_software.html

   To build:
     gcc -c -Wall -O2 [-DDEBUG] spamplot.c
     gcc -o spamplot spamplot.o -lgd -lfreetype -lpng -lz -lm
         or
     gcc -o spamplot spamplot.o \
         /usr/local/lib/libgd.a /usr/local/lib/libfreetype.a \
         /usr/local/lib/libpng/libpng.a /usr/local/lib/zlib/libz.a -lm
    
   History:
     20081230 - 0.1 - initial mockup (canvas, fake data; no labels, no ticks)
     20081231 - 0.2 - month/year labels and boxes, ticks, spam-count labels
     20090101 - 1.0 - title, real data, alternating main-canvas background
     20090104 - 1.1 - cmdline opts, usage, margin tweak, monthly stats, border
     // TODO:  nuke DEBUG -> stdout.png
     // TODO:  isatty() -> usage

     // TODO?  add input/output-file options
     // TODO?  reduce size of month-labels?  change font?  smallcaps?
     // TODO?  add fiducial 10's or 25's (maybe dotted or very pale green)?

  ---------------------------------------------------------------------------

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.
 
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  ---------------------------------------------------------------------------*/

#define DEBUG   // GRR TEMP

#define VERSION  "1.1 of 4 January 2009"

#include <stdio.h>
#include <stdlib.h>     /* atol() */
#include <ctype.h>      /* isdigit() */
#include <string.h>     /* strrchr() */

#define _XOPEN_SOURCE   /* required for older versions of glibc (man page) */
#define __USE_XOPEN     /* required for Slackware 10.0 version of glibc */
#include <time.h>       /* strptime() */

#include "gd.h"

typedef unsigned char  uch;
typedef unsigned long  ulg;

typedef struct {
    uch r;
    uch g;
    uch b;
} RGB;

typedef struct {
    int yday;      // day-offset from beginning of start_year
    int num_spam;  // our main vertical data point of interest...
    int month;     // index into month_colors[]
} SpamData;

static char *month_names[12] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun",
                                "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};

static int month_days_nonleap[12] = {31,28,31,30,31,30,31,31,30,31,30,31};
static int month_days_leap[12]    = {31,29,31,30,31,30,31,31,30,31,30,31};

static RGB month_colors[12] = {
    { 170,  41,  68 },  // { 0xaa, 0x29, 0x44 },   // dark red
    { 205,   0,   0 },  // { 0xcd, 0x00, 0x00 },   // bright red
    { 221,  75,   0 },  // { 0xdd, 0x4b, 0x00 },   // red-orange
    { 221, 133,   0 },  // { 0xdd, 0x85, 0x00 },   // orange
    { 205, 205,   0 },  // { 0xcd, 0xcd, 0x00 },   // yellow
    { 139, 160,   0 },  // { 0x8b, 0xa0, 0x00 },   // tan/olive
    {  67, 132,   0 },  // { 0x43, 0x84, 0x00 },   // green
    {  42,  95,  54 },  // { 0x2a, 0x5f, 0x36 },   // dark green
    {   0,  77, 190 },  // { 0x00, 0x4d, 0xbe },   // blue
    {  69,   0, 186 },  // { 0x45, 0x00, 0xba },   // blue-purple
    { 101,  26, 155 },  // { 0x65, 0x1a, 0x9b },   // purple-magenta
    { 139,   0, 139 }   // { 0x8b, 0x00, 0x8b }    // magenta
};

const double PI = 3.14159265358979323846;

const int maxDataPoints = 8100;  // big enough for 22 years and a bit
const int spamMax    = 800;      // big enough so far except for 20041211 (1059)
const int topMargin  =  50;
const int botMargin  =  17;
const int sideMargin =  50;

int main(int argc, char *argv[])
{
#ifdef DEBUG
    const char *outfn = "spamplot-debug.png";
    FILE *out;
#endif
    gdImagePtr im;
    SpamData spamdata[maxDataPoints];  // should be in order, oldest to newest
    int num_datapoints = 0;
    int white, light_gray, medium_gray, dark_gray, black,
        pale_yellow1, pale_yellow2, bright_grn;
    int gray, pale_yellow;
    int img_height, img_width;
    int month_gdColors[12];
    int *month_days;
    int num_days = 0, year_days;
    int prev_month = -1, num_spam_for_month = 0;
    int brect[8];        // font bounding box
    int font_height, str_width, max_str_width, gray_box_height;
    int xoff, xoff_start, yoff;
    int year = 0, start_year = 0, end_year = 0, max_year = 0;
    int c, c1, j;
    int error = 0, help = 0, verbose = 0;
    char *prognam, *fullnam=argv[0];
    char *q, spamline[256];
    char *errorStr, *font_name, label_str[128];


    /*-----------------------------------------------------------------------*
        Parse command-line options.
     *-----------------------------------------------------------------------*/

    if ((prognam = strrchr(fullnam, '/')) == NULL)
        prognam = fullnam;
    else
        ++prognam;

    while (--argc > 0  &&  (*++argv)[0] == '-') {
        while ((c = *++(argv[0]))) {	/* argv[0] is no longer program name */
            switch (c) {

                case 'e':               /* end-year */
                    c1 = argv[0][1];	/* get next char w/o changing ptr */
                    if ((c1 >= '0') && (c1 <= '9')) {
                        end_year = atol(argv[0]+1); /* ptr to digit string */
                        argv[0][1] = '\0';            /* overwrite digit */
                    } else {
                        --argc;
                        ++argv;
                        if (argc > 0 && argv[0][0] >= '0' && argv[0][0] <= '9')
                        {
                            end_year = atol(argv[0]); /* ptr to digit str */
                            while (*++(argv[0]))
                                ;
                            --(argv[0]);
                        } else {
                            fprintf(stderr, "error:  `%c' option requires "
                              "(numerical) year argument\n", c);
                            ++error;
                        }
                    }
                    break;

                case 'h':		/* just print help message */
                    help = 1;
                    break;

                case 's':               /* start-year */
                    c1 = argv[0][1];	/* get next char w/o changing ptr */
                    if ((c1 >= '0') && (c1 <= '9')) {
                        start_year = atol(argv[0]+1); /* ptr to digit string */
                        argv[0][1] = '\0';            /* overwrite digit */
                    } else {
                        --argc;
                        ++argv;
                        if (argc > 0 && argv[0][0] >= '0' && argv[0][0] <= '9')
                        {
                            start_year = atol(argv[0]); /* ptr to digit str */
                            while (*++(argv[0]))
                                ;
                            --(argv[0]);
                        } else {
                            fprintf(stderr, "error:  `%c' option requires "
                              "(numerical) year argument\n", c);
                            ++error;
                        }
                    }
                    break;

                case 'v':               /* verbosity level */
                    ++verbose;
                    break;


                default:
                    fprintf(stderr,
                      "error:  `%c' is not a valid option\n", c);
                    ++error;
                    break;
            }  /* end switch over option types */
        }  /* end while-loop over options within single argument */
    }  /* end while-loop over option-arguments */

    /*-----------------------------------------------------------------------*
        Print version/usage info if no input data.
     *-----------------------------------------------------------------------*/

    if (help || error) {
        fprintf(stderr,
"spamplot, version " VERSION ", by Greg Roelofs.\n"
"\n"
"usage:  %s [-h] [-v] [-s YYYY] [-e YYYY]\n"
"\n"
"  -h       display help/usage screen (what you're reading)\n"
"  -v       verbose mode\n"
"  -s YYYY  start_year, i.e., first year to be graphed\n"
"  -e YYYY  end_year, i.e., last year to be graphed\n"
"\n"
"Read spam data from stdin (format \"2003 Aug 19:  727\", one day per line,\n"
"oldest first; comments starting with '#' will be ignored) and plot the\n"
"specified year range in PNG format.  By default, the image will be written\n"
"to stdout.  If either start_year or end_year is unspecified, the first or\n"
"last year (respectively) present in the input data will be used instead;\n"
"by default, all years are plotted.  In verbose mode, also print monthly\n"
"statistics to stderr.\n",
          prognam);
        fflush(stderr);
        return error? 1 : 0;
    }

    /*-----------------------------------------------------------------------*
        Read spam data from stdin.
     *-----------------------------------------------------------------------*/

//  if ((infile = fopen(argv[1], "rb")) == NULL) {
//      fprintf(stderr, "ERROR:  can't open %s.\n", argv[1]);
//      return 6;
//  }

    do {   // loop over valid lines in input data

        // "2002 Jan 31:  13"
        // "2004 Dec  1:  358	# includes 1 from /mail/spam/spam.20050102.bz2"
        do {
            q = fgets(spamline, sizeof(spamline), stdin);  // NULL on error/EOF
        } while (q && (!isdigit(spamline[0]) || spamline[11] != ':'));

        if (q) {
            struct tm tm;
            // NULL on error, else points to char after last used:
            char *p = strptime(spamline, "%Y %b %d: ", &tm);

            if (p) {
                int num_spam;

                year = tm.tm_year + 1900;
                if (start_year == 0)
                    start_year = year;
                if (max_year < year)
                    max_year = year;

                if (sscanf(p, "%d", &num_spam) != 1) {
                    fprintf(stderr, "spamplot ERROR:  sscanf() failed to "
                      "convert \"%s\" at offset %d\n", spamline,
                      (int)(p - spamline));
                    fflush(stderr);
                    continue;
                }
                //printf("GRR DEBUG:  read \"%4d-%02d-%02d\" (yearday %d); next char = '%c'; num_spam = %d\n", year, tm.tm_mon + 1, tm.tm_mday, tm.tm_yday, *p, num_spam); fflush(stdout);

                // now stuff into data struct
                // want:  num_spam, ydays since beginning of start_year
                // or just convert into yeardays since start => no need for year
                if (year >= start_year && (end_year == 0 || year <= end_year)) {
                    int leapdays = 0;

                    // if assume input is in order, could replace this loop
                    // with simple tracking of tm_ydays seen so far...  FIXME?
                    for (j = start_year;  j < year;  ++j) {
                        if (j % 4 == 0)
                            ++leapdays;
                    }
                    // yday is now the offset from the first pixel column of
                    // the main canvas, i.e., add it to sideMargin+1 to get "x"
                    // (tm.tm_yday = [0..364 or 365]):
                    // FIXME:  this calculation ASSUMES first data point is
                    //         oldest...
                    spamdata[num_datapoints].yday =
                      (year - start_year)*365 + leapdays + tm.tm_yday;
                    spamdata[num_datapoints].num_spam =
                      (num_spam <= spamMax)? num_spam : spamMax;
                    spamdata[num_datapoints].month = tm.tm_mon;  // [0..11]
                    ++num_datapoints;
                    if (verbose) {
                        if (prev_month == tm.tm_mon) {
                            num_spam_for_month += num_spam;
                        } else {
                            // month just changed => print prev. month's totals
                            if (prev_month >= 0) {
                                // following will be incorrect if prev_month ==
                                // December (wrong year), but doesn't matter
                                // since only February is affected:
                                month_days = (year & 3)?
                                    month_days_nonleap : month_days_leap;
                                // print average daily spam in float format
                                fprintf(stderr, "  %d.%02d:  %7.2f\n",
                                  (prev_month == 11)? year-1 : year,
                                  prev_month+1, (float)
                                  num_spam_for_month / month_days[prev_month]);
                            }
                            prev_month = tm.tm_mon;
                            num_spam_for_month = num_spam;
                        }
                    }
                    if (num_datapoints > maxDataPoints)
                        break;
                }

            } else {
                fprintf(stderr, "spamplot ERROR:  strptime() failed to convert "
                  "\"%s\"\n", spamline);
                fflush(stderr);
                continue;
            }

        } else {
            // hit EOF or other error:  bail if no data?
            //printf("GRR DEBUG:  end of road on stdin\n"); fflush(stdout);
            break;
        }

    } while (1);

    // 2557 data points for 2002-2008, inclusive (== 365*7 + 2)
    // 4810 data points for Nov 1995-2008, inclusive (== 61 + 365*13 + 4)
    //printf("GRR DEBUG:  read %d valid data points\n", num_datapoints); fflush(stdout);

    /* print final month's average spams-per-day value (assuming had data) */
    if (verbose && year > 0) {
        // note that prev_month == tm.tm_mon here, so year variable is correct
        if (prev_month >= 0) {
            month_days = (year & 3)?
                month_days_nonleap : month_days_leap;
            fprintf(stderr, "  %d.%02d:  %7.2f\n", year, prev_month+1,
              (float)num_spam_for_month / month_days[prev_month]);
            fflush(stderr);
        }
    }

    if (end_year == 0)   // not specified by user:  use max seen in input data
        end_year = max_year;

    /* check for logic errors */
    if (start_year < 1980) {
        fprintf(stderr, "spamplot ERROR:  start_year = %d\n", start_year);
        fflush(stderr);
        return 1;
    }
    if (end_year < start_year) {
        fprintf(stderr, "spamplot ERROR:  end_year = %d < %d = start_year\n",
           end_year, start_year);
        fflush(stderr);
        return 1;
    }

    /* figure out width of graph */
    for (year = start_year;  year <= end_year;  ++year) {
        /* correct except for 1900 (no spam), 2100, 2200, 2300 (likely dead): */
        num_days += (year & 3)? 365 : 366;
    }

    /*-----------------------------------------------------------------------*
        Create gd image and allocate colors.
     *-----------------------------------------------------------------------*/

    /* to avoid having spamMax fiducial line right next to top border, we let
     * top border BE spamMax fiducial line, i.e., number of _interior_ pixels
     * (vertically) is spamMax-1 => bottom border = 0, top border = spamMax */
    img_width = sideMargin + 1 + num_days + 1 + sideMargin;
    img_height = topMargin + 1 + spamMax-1 + 1 + 92 + botMargin;
    im = gdImageCreate(img_width, img_height);

    white        = gdImageColorResolve(im, 255, 255, 255);  // 1st = bkgd color
    light_gray   = gdImageColorResolve(im, 204, 204, 204);
    medium_gray  = gdImageColorResolve(im, 188, 188, 188);  // was 153, 153, 153
    dark_gray    = gdImageColorResolve(im, 102, 102, 102);
    black        = gdImageColorResolve(im,   0,   0,   0);
    pale_yellow1 = gdImageColorResolve(im, 255, 250, 205);
    pale_yellow2 = gdImageColorResolve(im, 235, 230, 188);  // HSV: V: 100 -> 92
    bright_grn   = gdImageColorResolve(im,   0, 205,   0);
    for (j = 0;  j < 12;  ++j) {
        month_gdColors[j] = gdImageColorResolve(im, month_colors[j].r,
                                                    month_colors[j].g,
                                                    month_colors[j].b);
    }

    /*-----------------------------------------------------------------------*
        Draw main-canvas border and image border.
     *-----------------------------------------------------------------------*/

    /* bounding box for main canvas */
    gdImageLine(im, sideMargin,                topMargin,            // top
                    sideMargin + num_days + 1, topMargin,
                    dark_gray);
    gdImageLine(im, sideMargin + num_days + 1, topMargin,            // right
                    sideMargin + num_days + 1, topMargin + spamMax,
                    dark_gray);
    gdImageLine(im, sideMargin + num_days + 1, topMargin + spamMax,  // bottom
                    sideMargin,                topMargin + spamMax,
                    dark_gray);
    gdImageLine(im, sideMargin,                topMargin + spamMax,  // left
                    sideMargin,                topMargin,
                    dark_gray);

    /* background color for main canvas */
//  gdImageFilledRectangle(im, sideMargin + 1,        topMargin + 1,
//                             sideMargin + num_days, topMargin + spamMax-1,
//                             pale_yellow);

    /* black border inset two pixels from edges */
    gdImageLine(im, 2,             2,               // top
                    img_width - 3, 2,
                    black);
    gdImageLine(im, img_width - 3, 2,               // right
                    img_width - 3, img_height - 3,
                    black);
    gdImageLine(im, img_width - 3, img_height - 3,  // bottom
                    2,             img_height - 3,
                    black);
    gdImageLine(im, 2,             img_height - 3,  // left
                    2,             2,
                    black);

    /* second line inset one additional pixel from first */
    gdImageLine(im, 3,             3,               // top
                    img_width - 4, 3,
                    black);
    gdImageLine(im, img_width - 4, 3,               // right
                    img_width - 4, img_height - 4,
                    black);
    gdImageLine(im, img_width - 4, img_height - 4,  // bottom
                    3,             img_height - 4,
                    black);
    gdImageLine(im, 3,             img_height - 4,  // left
                    3,             3,
                    black);

    /*-----------------------------------------------------------------------*
        Draw ticks and numeric labels on both vertical axes.
     *-----------------------------------------------------------------------*/

    /* DEFAULT_FONTPATH already includes /usr/X11R6/lib/X11/fonts/TTF, but
     * if necessary, also can setenv GDFONTPATH appropriately at runtime */

    /* ticks on right and left sides (2-pixel for 10's, 4-pixel for 100's) */
    font_name = "timesbd";
    for (j = 0;  j <= spamMax;  j += 10) {
        int ticklen = 2;

        if (j % 100 == 0) {
            ticklen = 4;
            snprintf(label_str, sizeof(label_str), "%d", j);
            errorStr = gdImageStringFT(NULL, brect, black, font_name, 14.0, 0.0,
                                       100, 100, label_str);
            /* left-side axis labels */
            errorStr = gdImageStringFT(im, brect, black, font_name, 14.0, 0.0,
              sideMargin - ticklen - 2 - (brect[2]-brect[0]),  // 0-deg width
              topMargin + spamMax - j + (brect[1]-brect[7])/2, // half-height
              label_str);
            /* right-side axis labels */
            errorStr = gdImageStringFT(im, brect, black, font_name, 14.0, 0.0,
              sideMargin + num_days + 1 + ticklen + 2,         // left-justified
              topMargin + spamMax - j + (brect[1]-brect[7])/2, // half-height
              label_str);
        }

        gdImageLine(im, sideMargin - ticklen, topMargin + spamMax - j,
                        sideMargin - 1,       topMargin + spamMax - j,
                        dark_gray);
        gdImageLine(im, sideMargin + num_days + 1 + 1, topMargin + spamMax - j,
                        sideMargin + num_days + 1 + ticklen,
                        topMargin + spamMax - j, dark_gray);
    }

    /* initialize some basic font/label-related values... */

    /* bounding rectangle of string without descenders => "true" font height */
    font_name = "Cyberbit";
    errorStr = gdImageStringFT(NULL, brect, black, font_name, 14.0, PI/2,
                               100, 100, month_names[2]);  // "Mar"
    max_str_width = (brect[1]-brect[3]); // good approximation, not true max
    font_height = (brect[0]-brect[6]);   // 90-deg height

    xoff = sideMargin + 1;  // immediately inside border

    // font_height*3/2:  height estimate of 20pt bold relative to 14pt normal
    gray_box_height = 6 + max_str_width + 10 + font_height*3/2 + 16;

    /*
     * bounding rectangle:
     *
     *   0  lower left corner, X position
     *   1  lower left corner, Y position
     *   2  lower right corner, X position    (when string viewed horizontally)
     *   3  lower right corner, Y position
     *   4  upper right corner, X position
     *   5  upper right corner, Y position
     *   6  upper left corner, X position
     *   7  upper left corner, Y position
     *
     * Note that "lower left," "lower right," etc., are relative to the text
     * _when viewed horizontally_, but "X" and "Y" are relative to the _canvas_
     * (i.e., "absolute" coordinates).  Also note that "X" and "Y" refer to
     * the location of the font's _baseline_ (not lowest descender) and left
     * edge of the first character, which is NOT the same as the bounding box!
     * Thus for "Jan" at 100,100:
     *
     *   0 degrees / 0 radians:     brect = 98,101  126,101  126,86  98,86
     *                              half-height = (brect[1]-brect[7])/2
     *
     *  90 degrees / PI/2 radians:  brect = 101,101  101,74  86,74  86,101
     *                              half-height = (brect[0]-brect[6])/2

GRR DEBUG:  "Jan" brect = 75,887  75,857  59,857  59,887
GRR DEBUG:  "Feb" brect = 105,889  105,857  89,857  89,889
GRR DEBUG:  "Mar" brect = 134,891  134,856  118,856  118,891
GRR DEBUG:  "Apr" brect = 172,890  172,857  151,857  151,890   XV:
GRR DEBUG:  "May" brect = 201,893  201,856  181,856  181,893   21x38 at 181,856
GRR DEBUG:  "Jun" brect = 227,888  227,857  210,857  210,888
GRR DEBUG:  "Jul" brect = 257,883  257,857  240,857  240,883
GRR DEBUG:  "Aug" brect = 293,892  293,857  273,857  273,892
GRR DEBUG:  "Sep" brect = 325,889  325,857  304,857  304,889
GRR DEBUG:  "Oct" brect = 348,887  348,856  332,856  332,887
GRR DEBUG:  "Nov" brect = 379,892  379,857  363,857  363,892
GRR DEBUG:  "Dec" brect = 409,892  409,857  393,857  393,892

    // times, timesbd
    // VeraSe, VeraSeBd
    // Cyberbit
    // luxi...

// GRR DEBUG ONLY:
        errorStr = gdImageStringFT(im, brect, black, font_name, 14.0, PI/2,
          sideMargin,
          topMargin + spamMax,
          month_names[7]);  // Aug
printf("\nGRR EXTRA DEBUG:  \"%s\" brect = %d,%d  %d,%d  %d,%d  %d,%d\n", month_names[7], brect[0], brect[1], brect[2], brect[3], brect[4], brect[5], brect[6], brect[7]); fflush(stdout);

GRR EXTRA DEBUG:  "Aug" brect = 55,851  55,816  35,816  35,851

     */


    /*-----------------------------------------------------------------------*
        Draw title.
     *-----------------------------------------------------------------------*/

    font_name = "timesbd";
    snprintf(label_str, sizeof(label_str), "Sbam e-mails ber dab. %d - %d",
      start_year, end_year);  // want no descenders for positioning calc!
    errorStr = gdImageStringFT(NULL, brect, black, font_name, 20.0, 0.0,
                               100, 100, label_str);
//printf("\nGRR EXTRA DEBUG:  \"%s\" brect = %d,%d  %d,%d  %d,%d  %d,%d\n", label_str, brect[0], brect[1], brect[2], brect[3], brect[4], brect[5], brect[6], brect[7]); fflush(stdout);
// GRR EXTRA DEBUG:  "Sbam e-mails ber dab, 2002 - 2005" brect = 99,106  482,106  482,80  99,80
// GRR EXTRA DEBUG:  "Sbam e-mails ber dab. 2002 - 2005" brect = 99,101  482,101  482,80  99,80
// XV, non-descender:  376x18 at 591,17 => 591,34  966,34  966,17  591,17
//        gd translated 99,80 to 591,17 => 591,38  974,38  974,17  591,17:
//                                         4 pixels taller, 8 pixels wider! wtf?
    snprintf(label_str, sizeof(label_str), "Spam e-mails per day, %d - %d",
      start_year, end_year);
    errorStr = gdImageStringFT(im, brect, black, font_name, 20.0, 0.0,
      xoff + num_days/2 - (brect[2]-brect[0])/2,  // 0-deg half-width
      topMargin/2 + (brect[1]-brect[7])/2,        // 0-deg half-height, non-des.
      label_str);
    if (errorStr) {
        fprintf(stderr, "spamplot ERROR:  gdImageStringFT(\"%s\") failed:\n"
          "  \"%s\"\n", label_str, errorStr);
        fflush(stderr);
        // ...but keep going anyway
    }


    /*-----------------------------------------------------------------------*
        Draw month and year labels along bottom axis and background colors
        for main canvas.
     *-----------------------------------------------------------------------*/

    for (year = start_year;  year <= end_year;  ++year) {
        xoff_start = xoff;  // save for later (centered year-string)

        /* correct except for 1900 (no spam), 2100, 2200, 2300 (likely dead): */
        if ((year & 3)) {
            year_days = 365;
            month_days = month_days_nonleap;
        } else {
            year_days = 366;
            month_days = month_days_leap;
        }

        /* alternating background color for month/year labels at bottom */
        gray = (year & 1)? light_gray : medium_gray;
        gdImageFilledRectangle(im,
          xoff,                 topMargin + spamMax + 1,
          xoff + year_days - 1, topMargin + spamMax + 1 + gray_box_height - 1,
          gray);

        /* alternating background color for background of main canvas, too */
        pale_yellow = (year & 1)? pale_yellow1 : pale_yellow2;
        gdImageFilledRectangle(im,
          xoff,                 topMargin + 1,
          xoff + year_days - 1, topMargin + spamMax - 1,
          pale_yellow);

        /* loop over months for single year */
        font_name = "Cyberbit";
        for (j = 0;  j < 12;  ++j) {
            /* first get bounding rectangle so we can "right"-justify string */
            errorStr = gdImageStringFT(NULL, brect, black, font_name, 14.0,
                                       PI/2, 100, 100, month_names[j]);
            str_width = (brect[1]-brect[3]);  // for "right"-justification
            if (max_str_width < str_width)
                max_str_width = str_width;
            /* now actually render it */
            errorStr = gdImageStringFT(im, brect, black, font_name, 14.0, PI/2,
              xoff + month_days[j]/2 + font_height/2, // baseline (ignore desc.)
              topMargin + spamMax + 6 + str_width,    // right (top) justified
              month_names[j]);
//printf("GRR DEBUG:  \"%s\" brect = %d,%d  %d,%d  %d,%d  %d,%d\n", month_names[j], brect[0], brect[1], brect[2], brect[3], brect[4], brect[5], brect[6], brect[7]); fflush(stdout);
            if (errorStr) {
                fprintf(stderr, "spamplot ERROR:  gdImageStringFT(\"%s\") "
                  "failed:\n  \"%s\"\n", month_names[j], errorStr);
                fflush(stderr);
            }
            xoff += month_days[j];
        }

        /* At this point, xoff is full "year-width" beyond xoff_start, so half
         * of difference (or average of two) gives horizontal center for year
         * string.  Again, start by getting bounding rectangle. */
        font_name = "timesbd";
        snprintf(label_str, sizeof(label_str), "%d", year);
        errorStr = gdImageStringFT(NULL, brect, black, font_name, 20.0, 0.0,
                                   100, 100, label_str);
        errorStr = gdImageStringFT(im, brect, black, font_name, 20.0, 0.0,
          (xoff_start + xoff)/2 - (brect[2]-brect[0])/2,  // 0-deg half-width
          topMargin + spamMax + 6 + max_str_width + 10 + (brect[1]-brect[7]),
          label_str);                                  // ^^^ 0-deg full-height
        if (errorStr) {
            fprintf(stderr, "spamplot ERROR:  gdImageStringFT(\"%s\") failed:\n"
              "  \"%s\"\n", label_str, errorStr);
            fflush(stderr);
        }

    }

    /*-----------------------------------------------------------------------*
        Draw fiducial lines (spam levels) on main canvas.
     *-----------------------------------------------------------------------*/

    for (j = 100;  j < spamMax;  j += 100) {
        gdImageLine(im, sideMargin + 1,        topMargin + spamMax - j,
                        sideMargin + num_days, topMargin + spamMax - j,
                        bright_grn);
    }


    /*-----------------------------------------------------------------------*
        Plot spam data with a different color for each month.
     *-----------------------------------------------------------------------*/

    xoff = sideMargin + 1;           // first day:  immediately inside border
    yoff = topMargin + spamMax - 1;  // immediately inside bottom border
    for (j = 0;  j < num_datapoints;  ++j) {
        if (spamdata[j].num_spam > 0) {
            gdImageLine(im,
              xoff + spamdata[j].yday, yoff,
              xoff + spamdata[j].yday, yoff - (spamdata[j].num_spam - 1),
              month_gdColors[ spamdata[j].month ]);
        }
    }

    /*-----------------------------------------------------------------------*
        Write gd image to disk (or stdout) in PNG format.
     *-----------------------------------------------------------------------*/

#ifdef DEBUG
    if ((out = fopen(outfn, "wb")) == NULL) {
        fprintf(stderr,
          "spamplot ERROR:  can't open output file %s for writing\n", outfn);
        return 3;
    }
    gdImagePng(im, out);
    fclose(out);
#else
    gdImagePng(im, stdout);
#endif

    /*-----------------------------------------------------------------------*
        Clean up and quit.
     *-----------------------------------------------------------------------*/

    gdImageDestroy(im);  // destroy the image in memory

    return 0;
}
